/*
    This file is part of hat.

    hat is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    hat is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with hat.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>

#include "amoeba.h"

#define NMAX 100
#define ALPHA 1.0
#define BETA 0.5
#define GAMMA 2.0

#define SIMPLEX_SUM(simplex, sum)                             \
{                                                             \
	for (j = 0; j < COUNT; j++)                           \
		for (i = 0, sum[j] = 0.0; i < COUNT + 1; i++) \
			sum[j] += simplex[i][j];              \
}

double
amoeba_try(double simplex[COUNT + 1][COUNT], double y[COUNT + 1],
           double *sum, double (*f)(), void *data, int ihi, double fac)
{
	int j;
	double fac1, fac2, y_try;
	double try[COUNT];

	fac1 = (1.0 - fac) / COUNT;
	fac2 = fac1 - fac;

	for (j = 0; j < COUNT; j++)
		try[j] = sum[j] * fac1 - simplex[ihi][j] * fac2;

	y_try = (*f)(try, data);
	if (y_try < y[ihi]) {
		y[ihi] = y_try;
		for (j = 0; j < COUNT; j++) {
			sum[j] += try[j] - simplex[ihi][j];
			simplex[ihi][j] = try[j];
		}
	}
	return y_try;
}

int
amoeba(double simplex[COUNT + 1][COUNT],
       double toll, double (*f)(), void *data)
{
	int i, j;
	int ilo, ihi, inhi; /* Low hi and next-hi points */
	int iter;

	double y_try, y_save, r_toll;
	double tmp_div;
	double y[COUNT + 1];
	double psum[COUNT];

	iter = 0;
	for (i = 0; i < COUNT + 1; i++)
		y[i] = f(simplex[i], data);

	SIMPLEX_SUM(simplex, psum);

	while (1) { /* Main loop */
		ilo = 1; /* Determine ilo, ihi and inhi */
		ihi = y[1] > y [2] ? (inhi = 2, 1) : (inhi = 1, 2);
		for (i = 0; i < COUNT + 1; i++) {
			if (y[i] < y[ilo])
				ilo = i;
			if (y[i] > y[ihi]) {
				inhi = ihi;
				ihi = i;
			} else if
				(y[i] > y[inhi])
				if (i != ihi)
					inhi = i;
		}

		tmp_div = fabs(y[ihi]) + fabs(y[ilo]);
		if (tmp_div < 1.0e-11)
			tmp_div = 1.0e-11; /* Prevent dividing by zero */

		r_toll = 2.0 * fabs(y[ihi] - y[ilo]) / tmp_div;
		if ((r_toll < toll) || (iter >= NMAX))
			break; /* Tollerance reached */

		y_try = amoeba_try(simplex, y, psum, f, data, ihi, -ALPHA);
		iter++;

		if (y_try <= y[ilo]) 
			y_try = amoeba_try(simplex, y, psum,
                                           f, data, ihi, GAMMA);

		else if (y_try >= y[inhi]) {
			y_save = y[ihi]; 
			y_try = amoeba_try(simplex, y, psum, f, data, ihi, BETA);
			if (y_try >= y_save) { 
				for (i = 0; i < COUNT + 1; i++) {
					if (i != ilo) {
						for (j = 0; j < COUNT; j++) {
							psum[j] = simplex[i][j] + simplex[ilo][j];
							psum[j] *= 0.5;
							simplex[i][j] = psum[j];
						}
						y[i] = (*f)(psum, data);
					}
				}
				iter += COUNT;
				
				SIMPLEX_SUM(simplex, psum);
			}
		}
	}

	return ilo;
}

