/*
    This file is part of hat.

    hat is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    hat is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with hat.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>
#include <complex.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_min.h>

#include "fa_utils.h"

double
get_amplitude(double *buf, int count, double q)
{
        int j; 
        double amp;
        double complex z, z_aux;
	double complex twiddle, twiddle_0;

	twiddle_0 = cexp(2. * M_PI * I * (double complex)(q));

        z = 0.;
	twiddle = 1.;

        for (j = 0; j < count; j++) {

                z_aux = buf[j];
                if (isnan(z_aux))
                        continue;

                z_aux *= 1. - (double complex)cos(2. * M_PI * (double)j / (double)(count - 1));
                z_aux *= twiddle;
                z += z_aux;

		twiddle *= twiddle_0;
        }

        amp = cabs(z);

        return amp;
}


double
get_principal_line_guess(double *x, int count, double q_min, double q_max)
{
        int i;
        int max_idx;
        double max, max_p1, max_m1;
        double amp, amp_previous;
        double q;

        max_idx = -1.;
        max = 0.;
        max_p1 = NAN;
        max_m1 = NAN;
        amp_previous = NAN;

        q_min *= count;
        q_max *= count;

        if (q_min < 1)
                q_min = 1;

        if (q_max > count / 2 - 1)
                q_max = count / 2 - 1;

        for (i = q_min; i <= q_max; i++) {
		q = (double)i / (double)count;
		amp = get_amplitude(x, count, q);

                if (amp > max) {
                        max = amp;
                        max_idx = i;
                        max_m1 = amp_previous; /* Point before the maximum */
                }
                amp_previous = amp;
                if (i == max_idx + 1) /* Point after the maximum */
                        max_p1 = amp;
        }
        if (max_idx == q_min || max_idx == q_max || max_idx == -1)
                return NAN;

        q = max_idx / (double)count;
        return q;
}

struct Data {
        double *buf;       /* Data to be fitted */
        int turns;         /* Number of turns */
};

double
test_amplitude(double q, void *param)
{
        int count;
	double *buf;
	double amp;

	buf = ((struct Data *)param)->buf;
	count = ((struct Data *)param)->turns;

	amp = get_amplitude(buf, count, q);

	return -1. * amp;
}

double
get_principal_line_laskar(double *data, int turns, double guess)
{
	int status;
	int iter;
	gsl_min_fminimizer *minimizer;
	double lower_bound, upper_bound;
	gsl_function f;

	struct Data fit_data;

	fit_data.buf = data;
	fit_data.turns = turns;

	lower_bound = guess - 1. / (double)(turns);
	upper_bound = guess + 1. / (double)(turns); 

	if (test_amplitude(guess, &fit_data) >= test_amplitude(lower_bound, &fit_data) ||
	    test_amplitude(guess, &fit_data) >= test_amplitude(upper_bound, &fit_data)) {

		return NAN;
	}

	f.function = &test_amplitude;
	f.params = &fit_data;

	minimizer = gsl_min_fminimizer_alloc(gsl_min_fminimizer_brent);
	gsl_min_fminimizer_set(minimizer, &f, guess,
  	                       lower_bound, upper_bound);

	iter = 0;
	do {
		iter++;
		status = gsl_min_fminimizer_iterate(minimizer);
		if (status) {
			guess = NAN;
			break;
		}

		guess = gsl_min_fminimizer_x_minimum(minimizer);
		lower_bound = gsl_min_fminimizer_x_lower(minimizer);
		upper_bound = gsl_min_fminimizer_x_upper(minimizer);

		status = gsl_min_test_interval(lower_bound, upper_bound, 1e-6, 1e-6);
	} while (status == GSL_CONTINUE && iter < 10);

	gsl_min_fminimizer_free(minimizer);

	return guess;
}

double
get_principal_line(double *x, int count, double q_min, double q_max)
{
	double q;

	q = get_principal_line_guess(x, count, q_min, q_max);
	if (isnan(q))
		return NAN;

	q = get_principal_line_laskar(x, count, q);
	return q;
}

