/*
    This file is part of hat.

    hat is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    hat is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with hat.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <Python.h>
#include <numpy/arrayobject.h>
#include <complex.h>
#include <stdio.h>

#include "fa_utils.h"
#include "damping.h"
#include "fa.h"

static PyObject*
find_dominant_line_naff_wrapper(PyObject *dummy, PyObject *args)
{
        PyObject *arg1 = NULL, *arg2 = NULL;
        PyObject *data = NULL, *freq_bound = NULL;

        double q;
        npy_intp *count;
        double *data_buf, *freq_bound_buf;

        if (!PyArg_ParseTuple(args, "OO", &arg1, &arg2))
                return NULL;

        data       = PyArray_FROM_OTF(arg1, NPY_DOUBLE, NPY_IN_ARRAY);
        freq_bound = PyArray_FROM_OTF(arg2, NPY_DOUBLE, NPY_IN_ARRAY);
        if (data == NULL || freq_bound == NULL)
                return NULL;

        count = PyArray_DIMS(data);   //number of dimensions
        data_buf       = (double *)PyArray_DATA(data);
        freq_bound_buf = (double *)PyArray_DATA(freq_bound);

        q = get_principal_line(data_buf, (int)count[0],
                               freq_bound_buf[0], freq_bound_buf[1]);

        Py_DECREF(data);
        Py_DECREF(freq_bound);
        return Py_BuildValue("d", q);
}

static PyObject*
eval_line_naff_wrapper(PyObject *dummy, PyObject *args)
{
        PyObject *arg1 = NULL, *arg2 = NULL;
        PyObject *data = NULL;
	PyObject *z_obj;

        double q;
	double complex z;
        double *data_buf;
        npy_intp *count;

        if (!PyArg_ParseTuple(args, "OO", &arg1, &arg2))
                return NULL;

        q = PyFloat_AsDouble(arg2);

        data = PyArray_FROM_OTF(arg1, NPY_DOUBLE, NPY_IN_ARRAY);
        if (data == NULL)
                return NULL;

        count = PyArray_DIMS(data);   //number of dimensions
        data_buf       = (double *)PyArray_DATA(data);

        z = ortho_projection(data_buf, (int)count[0], q);
	z_obj = PyComplex_FromDoubles(creal(z), cimag(z));

        Py_DECREF(data);

        return z_obj;
}

static PyObject*
eval_damping_wrapper(PyObject *dummy, PyObject *args)
{
        PyObject *arg1 = NULL, *arg2 = NULL, *arg3 = NULL;
        PyObject *data = NULL;

        double q, phase, damping;
        double *data_buf;
        npy_intp *count;

        if (!PyArg_ParseTuple(args, "OOO", &arg1, &arg2, &arg3))
                return NULL;

        q     = PyFloat_AsDouble(arg2);
	phase = PyFloat_AsDouble(arg3);

        data = PyArray_FROM_OTF(arg1, NPY_DOUBLE, NPY_IN_ARRAY);
        if (data == NULL)
                return NULL;

        count = PyArray_DIMS(data);   //number of dimensions
        data_buf       = (double *)PyArray_DATA(data);

	damping = fit_damping(data_buf, (int)count[0], q, phase);

        Py_DECREF(data);

        return Py_BuildValue("d", damping);
}

static PyObject*
eval_damped_line_naff_wrapper(PyObject *dummy, PyObject *args)
{
        PyObject *arg1 = NULL, *arg2 = NULL, *arg3 = NULL;
        PyObject *data = NULL;

        double q, damping;
//	double re, im;
	double complex z;
	PyObject *z_obj;
        double *data_buf;
        npy_intp *count;

        if (!PyArg_ParseTuple(args, "OOO", &arg1, &arg2, &arg3))
                return NULL;

        q       = PyFloat_AsDouble(arg2);
        damping = PyFloat_AsDouble(arg3);

        data = PyArray_FROM_OTF(arg1, NPY_DOUBLE, NPY_IN_ARRAY);
        if (data == NULL)
                return NULL;

        count = PyArray_DIMS(data);   //number of dimensions
        data_buf       = (double *)PyArray_DATA(data);

        z = ortho_projection_damped(data_buf, (int)count[0], q, damping);

        z_obj = PyComplex_FromDoubles(creal(z), cimag(z));

        Py_DECREF(data);

        return z_obj;

        //Py_DECREF(data);

        //return Py_BuildValue("dd", re, im);
}

static PyObject*
purge_line_wrapper(PyObject *dummy, PyObject *args)
{
        PyObject *arg1 = NULL, *arg2 = NULL;
        PyObject *arg3 = NULL, *arg4 = NULL;
        PyObject *data = NULL;
	Py_complex z_obj;

        double q, damping;
	double complex z;
        double *data_buf;
        npy_intp *count;

        if (!PyArg_ParseTuple(args, "OOOO", &arg1, &arg2, &arg3, &arg4))
                return NULL;

        q       = PyFloat_AsDouble(arg2);
	z_obj   = PyComplex_AsCComplex(arg3);
	damping = PyFloat_AsDouble(arg4);

        data = PyArray_FROM_OTF(arg1, NPY_DOUBLE, NPY_IN_ARRAY);
        if (data == NULL)
                return NULL;

        count = PyArray_DIMS(data);   //number of dimensions
        data_buf       = (double *)PyArray_DATA(data);

	z = z_obj.real + I * z_obj.imag;

	purge_line(data_buf, (int)count[0], q, z, damping);

        Py_DECREF(data);

	Py_RETURN_NONE;
}

static struct PyMethodDef methods[] = {
        {"find_dominant_line_naff",
         find_dominant_line_naff_wrapper, METH_VARARGS,
         "Return the dominant spectral line in the"
         "given bound, using a Laskar like method"},

        {"eval_line_naff",
         eval_line_naff_wrapper, METH_VARARGS,
         "Evaluate amplitude and phase of a spectral line"},

        {"eval_damped_line_naff",
         eval_damped_line_naff_wrapper, METH_VARARGS,
         "Evaluate amplitude and phase of a spectral line in"
         "presence of damping"},

        {"eval_damping",
         eval_damping_wrapper, METH_VARARGS,
         "Evaluate damping"},

        {"purge_line",
         purge_line_wrapper, METH_VARARGS,
         "Remove a spectral line from the signal"},

        {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC
inithat(void)
{
	char *doc = "HAT is my personal attempt at an Harmonic Analysis Toolkit\n"
                    "Copyright Michele Carla' 2018";

    	(void)Py_InitModule3("hat", methods, doc);
    	import_array();
}

